import abc
import gc
from collections import defaultdict
from pathlib import Path
from typing import Any, Optional, Dict, Union, AnyStr, Tuple, Iterator

import numpy as np
import tensorflow as tf
from tqdm import tqdm

from cinnamon_core.core.data import FieldDict
from cinnamon_core.utility import logging_utility
from cinnamon_generic.components.callback import Callback, guard
from cinnamon_generic.components.metrics import Metric
from cinnamon_generic.components.model import Network
from cinnamon_generic.components.processor import Processor
from cinnamon_generic.utility.printing_utility import prettify_statistics


class TFNetwork(Network):

    @abc.abstractmethod
    def batch_loss(
            self,
            batch_x: Any,
            batch_y: Any,
            training: bool = False
    ) -> Tuple[Any, Any, Dict, Any, Dict]:
        """
        Computes model loss for given batch.

        Args:
            batch_x: batch input data in any model-compliant format
            batch_y: batch ground-truth data in any model-compliant format
            training: whether the model is in training mode or not

        Returns:
            The computed loss information for the current step (e.g., losses name and value)
        """

        pass

    def batch_train(
            self,
            batch_x: Any,
            batch_y: Any
    ) -> Tuple[Any, Any, Dict, Any, Dict, Any]:
        """
        Computes a training step given input data.

        Args:
            batch_x: batch input training data in any model-compliant format
            batch_y: batch ground-truth training data in any model-compliant format

        Returns:
            The following information is returned:
                - Total loss
                - Total true loss: total loss without any weighting applied (if no weighting is considered, it equals total loss)
                - Loss information: a dictionary with loss name as keys and loss value as values
                - Predictions: model raw output
                - Model additional info: optional intermediate or additional model raw output
                - Gradients: loss computed gradients
        """

        with tf.GradientTape() as tape:
            loss, \
                true_loss, \
                loss_info, \
                predictions, \
                model_additional_info = self.batch_loss(batch_x,
                                                        batch_y,
                                                        training=True)
        grads = tape.gradient(loss, self.model.trainable_variables)
        return loss, true_loss, loss_info, predictions, model_additional_info, grads

    @tf.function(experimental_relax_shapes=True)
    def batch_fit(
            self,
            batch_x: Any,
            batch_y: Any
    ) -> Tuple[Dict, Any, Dict]:
        """
        Computes a training step given input data.

        Args:
            batch_x: batch input training data in any model-compliant format
            batch_y: batch ground-truth training data in any model-compliant format

        Returns:
            The computed training information for the current step:
                - Loss information: a dictionary with loss name as keys and loss value as values
                - Predictions: model raw output
                - Model additional info: optional intermediate or additional model raw output
        """

        loss, \
            true_loss, \
            loss_info, \
            predictions, \
            model_additional_info, \
            grads = self.batch_train(batch_x=batch_x,
                                     batch_y=batch_y)

        self.optimizer.apply_gradients(zip(grads, self.model.trainable_variables))

        loss_info['loss'] = true_loss
        return loss_info, predictions, model_additional_info

    @tf.function(experimental_relax_shapes=True)
    def batch_predict(
            self,
            batch_x: Any,
    ) -> Tuple[Any, Dict]:
        """
        Computes model predictions for the given input batch.

        Args:
            batch_x: batch input training data in any model-compliant format

        Returns:
            - Predictions: model raw output
            - Model additional info: optional intermediate or additional model raw output
        """

        predictions, model_additional_info = self.model(batch_x,
                                                        training=False)
        return predictions, model_additional_info

    @tf.function(experimental_relax_shapes=True)
    def batch_evaluate(
            self,
            batch_x: Any,
            batch_y: Any
    ) -> Tuple[Any, Any, Dict, Any, Dict]:
        """
        Computes training loss for the given input batch without a issuing a gradient step.

        Args:
            batch_x: batch input training data in any model-compliant format
            batch_y: batch ground-truth training data in any model-compliant format

        Returns:
            The following information is returned:
                - Total loss
                - Total true loss: total loss without any weighting applied (if no weighting is considered, it equals total loss)
                - Loss information: a dictionary with loss name as keys and loss value as values
                - Predictions: model raw output
                - Model additional info: optional intermediate or additional model raw output
        """

        loss, \
            true_loss, \
            loss_info, \
            predictions, \
            model_additional_info = self.batch_loss(batch_x=batch_x,
                                                    batch_y=batch_y,
                                                    training=False)
        loss_info['loss'] = true_loss
        return loss, true_loss, loss_info, predictions, model_additional_info

    def save_model(
            self,
            filepath: Union[AnyStr, Path]
    ):
        """
        Serializes internal model's weights to filesystem.

        Args:
            filepath: path where to save model's weights.
        """

        self.model.save_weights(filepath=filepath.joinpath('weights.h5'))

    def load_model(
            self,
            filepath: Union[AnyStr, Path]
    ):
        """
        Loads internal model's weights from a serialized checkpoint stored in the filesystem.

        Args:
            filepath: path where the model serialized checkpoint is stored.
        """

        self.model.load_weights(filepath=filepath.joinpath('weights.h5'))

    @guard()
    def fit(
            self,
            train_data: FieldDict,
            val_data: Optional[FieldDict] = None,
            callbacks: Optional[Callback] = None,
            metrics: Optional[Metric] = None,
            model_processor: Optional[Processor] = None
    ) -> FieldDict:
        """
        Fits the model with given training and (optionally) validation data.

        Args:
            train_data: training data necessary for training the model
            val_data: validation data that can be used to regularize or monitor the training process
            callbacks: callbacks for custom execution flow and side effects
            metrics: metrics for quantitatively evaluate the training process
            model_processor: a ``Processor`` component that parses model predictions.

        Returns:
            A ``FieldDict`` storing training information
        """

        logging_utility.logger.info('Training started...')
        logging_utility.logger.info(f'Total steps: {train_data.steps}')

        training_info = {}
        for epoch in range(self.epochs):

            if self.model.stop_training:
                logging_utility.logger.info(f'Stopping training at epoch {epoch}')
                break

            if callbacks:
                callbacks.run(hookpoint='on_epoch_begin',
                              logs={'epochs': self.epochs, 'epoch': epoch})

            epoch_info = defaultdict(float)
            batch_idx = 0

            data_iterator: Iterator = train_data.iterator()
            with tqdm(total=train_data.steps, leave=True, position=0, desc=f'Training - Epoch {epoch}') as pbar:
                while batch_idx < train_data.steps:

                    if callbacks:
                        callbacks.run(hookpoint='on_batch_fit_begin',
                                      logs={'batch': batch_idx})

                    batch_info, _, model_additional_info = self.batch_fit(*next(data_iterator))
                    batch_info = {f'train_{key}': item.numpy() for key, item in batch_info.items()}

                    if callbacks:
                        callbacks.run(hookpoint='on_batch_fit_end',
                                      logs={'batch': batch_idx,
                                            'batch_info': batch_info,
                                            'model_additional_info': model_additional_info})

                    for key, item in batch_info.items():
                        epoch_info[key] += item

                    batch_idx += 1
                    pbar.update(1)

            epoch_info = {key: item / train_data.steps for key, item in epoch_info.items()}
            epoch_info['epoch'] = epoch

            if val_data is not None:
                val_info = self.evaluate(data=val_data,
                                         callbacks=callbacks,
                                         metrics=metrics,
                                         model_processor=model_processor,
                                         suffixes={'status': 'training'})
                val_info = val_info.to_value_dict()

                del val_info['predictions']
                if 'metrics' in val_info:
                    val_info = {**val_info, **val_info['metrics']}
                    del val_info['metrics']

                epoch_info = {**epoch_info, **{f'val_{key}': value for key, value in val_info.items()}}

            logging_utility.logger.info(f'\n{prettify_statistics(epoch_info)}')

            if callbacks:
                callbacks.run(hookpoint='on_epoch_end',
                              logs=epoch_info)

            for key, value in epoch_info.items():
                training_info.setdefault(key, []).append(value)

            # Garbage collect
            gc.collect()

        return FieldDict(training_info)

    @guard()
    def evaluate(
            self,
            data: FieldDict,
            callbacks: Optional[Callback] = None,
            metrics: Optional[Metric] = None,
            model_processor: Optional[Processor] = None,
            suffixes: Optional[Dict] = None
    ) -> FieldDict:
        """
        Evaluates a trained model on given data and computes model predictions on the same data.

        Args:
            data: data to evaluate the model on and compute predictions.
            callbacks: callbacks for custom execution flow and side effects
            metrics: metrics for quantitatively evaluate the training process
            model_processor: a ``Processor`` component that parses model predictions.
            suffixes: suffixes used to uniquely identify evaluation results on input data

        Returns:
            A ``FieldDict`` storing evaluation and prediction information
        """

        loss = defaultdict(float)
        predictions = []

        data_iterator: Iterator = data.iterator()
        for batch_idx in tqdm(range(data.steps), leave=True, position=0, desc='Evaluating'):

            if callbacks:
                callbacks.run(hookpoint='on_batch_evaluate_begin',
                              logs={'batch': batch_idx,
                                    'suffixes': suffixes})

            batch_loss, \
                true_batch_loss, \
                batch_loss_info, \
                batch_predictions, \
                model_additional_info = self.batch_evaluate(*next(data_iterator))

            batch_info = {key: item.numpy() for key, item in batch_loss_info.items()}

            if callbacks:
                callbacks.run(hookpoint='on_batch_evaluate_end',
                              logs={'batch': batch_idx,
                                    'batch_info': batch_info,
                                    'batch_loss': batch_loss,
                                    'true_batch_loss': true_batch_loss,
                                    'batch_predictions': batch_predictions,
                                    'model_additional_info': model_additional_info,
                                    'suffixes': suffixes})

            for key, item in batch_info.items():
                loss[key] += item

            if model_processor is not None:
                batch_predictions = model_processor.run(data=batch_predictions)
            else:
                batch_predictions = batch_predictions.numpy()
            predictions.extend(batch_predictions)

        predictions = np.array(predictions)
        loss = {key: item / data.steps for key, item in loss.items()}

        if 'output_iterator' not in data or metrics is None:
            metrics_info = {}
        else:
            ground_truth = np.concatenate([item for item in data.output_iterator()])
            metrics_info = metrics.run(y_pred=predictions, y_true=ground_truth, as_dict=True)

        return FieldDict({**loss, **{'metrics': metrics_info}, **{'predictions': predictions}})

    @guard()
    def predict(
            self,
            data: FieldDict,
            callbacks: Optional[Callback] = None,
            metrics: Optional[Metric] = None,
            model_processor: Optional[Processor] = None,
            suffixes: Optional[Dict] = None
    ) -> FieldDict:
        """
        Computes model predictions on the given data.

        Args:
            data: data to compute model predictions.
            callbacks: callbacks for custom execution flow and side effects
            metrics: metrics for quantitatively evaluate the training process
            model_processor: a ``Processor`` component that parses model predictions.
            suffixes: suffixes used to uniquely identify evaluation results on input data

        Returns:
            A ``FieldDict`` storing prediction information
        """

        predictions = []

        data_iterator: Iterator = data.input_iterator() if 'input_iterator' in data else data.iterator()
        for batch_idx in tqdm(range(data.steps), leave=True, position=0, desc='Predicting'):

            if callbacks:
                callbacks.run(hookpoint='on_batch_predict_begin',
                              logs={'batch': batch_idx,
                                    'suffixes': suffixes})

            batch_predictions, model_additional_info = self.batch_predict(next(data_iterator))
            if model_processor is not None:
                batch_predictions = model_processor.run(data=batch_predictions)
            else:
                batch_predictions = batch_predictions.numpy()
            predictions.extend(batch_predictions)

            if callbacks:
                callbacks.run(hookpoint='on_batch_predict_end',
                              logs={'batch': batch_idx,
                                    'batch_predictions': batch_predictions,
                                    'model_additional_info': model_additional_info,
                                    'suffixes': suffixes})

        predictions = np.array(predictions)
        if 'output_iterator' not in data or metrics is None:
            metrics_info = {}
        else:
            ground_truth = np.concatenate([item for item in data.output_iterator()])
            metrics_info = metrics.run(y_pred=predictions, y_true=ground_truth, as_dict=True)

        return FieldDict({**{'predictions': predictions}, **{'metrics': metrics_info}})



from typing import Type, Optional

from cinnamon_core.core.configuration import Configuration, C
from cinnamon_core.core.registry import register, Registry
from cinnamon_tf.components.callback import TFEarlyStopping


class TFEarlyStoppingConfig(Configuration):

    @classmethod
    def get_default(
            cls: Type[C]
    ) -> C:
        config = super().get_default()

        config.add(name='monitor',
                   value='val_loss',
                   type_hint=str,
                   is_required=True,
                   description='Metric/Loss to monitor for performing early stopping')
        config.add(name='patience',
                   value=5,
                   type_hint=int,
                   is_required=True,
                   allowed_range=lambda patience: patience > 0,
                   description='Number of epochs to wait for monitored value to improve before stopping training')
        config.add(name='baseline',
                   value=None,
                   type_hint=Optional[float],
                   description='Threshold value to consider for start monitoring')
        config.add(name='min_delta',
                   value=1e-06,
                   type_hint=float,
                   description='Minimum delta difference allowed in checking for new best monitor values')
        config.add(name='restore_best_weights',
                   value=True,
                   type_hint=bool,
                   description='If enabled, the best epoch weights will be restored before stopping training')
        config.add(name='mode',
                   value='min',
                   type_hint=str,
                   description='How to compute best monitor scores. '
                               'If the monitored value gets better as it diminishes or viceversa',
                   allowed_range=lambda mode: mode.casefold() in ['min', 'max'])

        return config


@register
def register_callbacks():
    Registry.add_and_bind(config_class=TFEarlyStoppingConfig,
                          component_class=TFEarlyStopping,
                          name='callback',
                          tags={'early_stopping'},
                          namespace='tf')



from typing import Type, Iterable

from cinnamon_core.core.configuration import Configuration, C
from cinnamon_core.core.registry import Registry, register
from cinnamon_tf.components.helper import TFHelper


class TFHelperConfig(Configuration):

    @classmethod
    def get_default(
            cls: Type[C]
    ) -> C:
        config = super().get_default()

        config.add(name='deterministic',
                   value=False,
                   type_hint=bool,
                   description='If enabled, tensorflow will run operations in deterministic mode.'
                               'Note that this behaviour may seldom raise CUDA-related errors.')
        config.add(name='limit_gpu_visibility',
                   value=True,
                   type_hint=bool,
                   description='If enabled, it forces Tensorflow gpu visibility to the specified devices only')
        config.add(name='gpu_indexes',
                   value=[0],
                   type_hint=Iterable[int],
                   description='List of gpu indexes to make available to Tensorflow')
        config.add(name='eager_execution',
                   value=False,
                   type_hint=bool,
                   description='Whether to execute in eager or in graph mode.')

        # If limit_gpu_visibility, make sure we have some indexes
        config.add_condition(name='gpu_visibility',
                             condition=lambda params: (params.limit_gpu_visibility and params.gpu_indexes is not None)
                                                      or not params.limit_gpu_visibility)

        return config


@register
def register_helpers():
    Registry.add_and_bind(config_class=TFHelperConfig,
                          component_class=TFHelper,
                          name='helper',
                          is_default=True,
                          namespace='tf')
    Registry.add_and_bind(config_class=TFHelperConfig,
                          config_constructor=TFHelperConfig.get_delta_class_copy,
                          config_kwargs={
                              'params': {
                                  'eager_execution': True
                              }
                          },
                          component_class=TFHelper,
                          name='helper',
                          tags={'eager'},
                          namespace='tf')



from typing import Optional, Dict

import numpy as np

from cinnamon_core.utility import logging_utility
from cinnamon_generic.components.callback import Callback


class TFEarlyStopping(Callback):
    """Stop training when a monitored quantity has stopped improving.
    Arguments:
        monitor: Quantity to be monitored.
        min_delta: Minimum change in the monitored quantity
            to qualify as an improvement, i.e. an absolute
            change of less than min_delta, will count as no
            improvement.
        patience: Number of epochs with no improvement
            after which training will be stopped.
        verbose: verbosity mode.
        mode: One of `{"auto", "min", "max"}`. In `min` mode,
            training will stop when the quantity
            monitored has stopped decreasing; in `max`
            mode it will stop when the quantity
            monitored has stopped increasing; in `auto`
            mode, the direction is automatically inferred
            from the name of the monitored quantity.
        baseline: Baseline value for the monitored quantity.
            Training will stop if the model doesn't show improvement over the
            baseline.
        restore_best_weights: Whether to restore model weights from
            the epoch with the best value of the monitored quantity.
            If False, the model weights obtained at the last step of
            training are used.
    Example:
    ```python
    callback = tf.keras.callbacks.EarlyStopping(monitor='val_loss', patience=3)
    # This callback will stop the training when there is no improvement in
    # the validation loss for three consecutive epochs.
    model.fit(data, labels, epochs=100, callbacks=[callback],
        validation_data=(val_data, val_labels))
    ```
    """

    def __init__(
            self,
            **kwargs
    ):
        super(TFEarlyStopping, self).__init__(**kwargs)

        self.wait: int = 0
        self.stopped_epoch: int = 0
        self.best_value: Optional[float] = None
        self.best_weights: Optional[int] = None

        if self.mode == 'min':
            self.monitor_op = np.less
            self.min_delta *= -1
            self.best_value = np.inf
        elif self.mode == 'max':
            self.monitor_op = np.greater
            self.min_delta *= 1
            self.best_value = -np.inf

    def reset(
            self
    ):
        """
        Resets the ``EarlyStopping`` internal state for component re-use.
        """

        self.wait = 0
        self.stopped_epoch = 0
        self.best_weights = None
        if self.baseline is not None:
            self.best_value = self.baseline
        else:
            self.best_value = np.Inf if self.monitor_op == np.less else -np.Inf

        self.component.model.stop_training = False

    def on_fit_begin(
            self,
            logs: Optional[Dict] = None
    ):
        # Allow instances to be re-used
        self.reset()

    def on_epoch_end(
            self,
            logs: Optional[Dict] = None
    ):
        current = self.get_monitor_value(logs)
        if current is None:
            return
        if self.monitor_op(current - self.min_delta, self.best_value):
            self.best_value = current
            self.wait = 0
            if self.restore_best_weights:
                self.best_weights = self.component.model.get_weights()
        else:
            self.wait += 1
            if self.wait >= self.patience:
                self.stopped_epoch = logs['epoch']
                self.component.model.stop_training = True
                if self.restore_best_weights:
                    logging_utility.logger.info('Restoring model weights from the end of the best epoch.')
                    self.component.model.set_weights(self.best_weights)

    def on_fit_end(
            self,
            logs: Optional[Dict] = None
    ):
        if self.stopped_epoch > 0:
            logging_utility.logger.info(f'Epoch {self.stopped_epoch + 1:.2f} early stopping')

        self.reset()

    def get_monitor_value(
            self,
            logs: Optional[Dict] = None
    ):
        logs = logs if logs is not None else {}
        monitor_value = logs.get(self.monitor)
        if monitor_value is None:
            logging_utility.logger.warning(
                f'Early stopping conditioned on metric {self.monitor} which is not available.'
                f' Available metrics are: {",".join(list(logs.keys()))}')
        return monitor_value



import os
from typing import Optional, Any

import tensorflow as tf
from tensorflow.python.keras import backend as K

from cinnamon_core.utility import logging_utility
from cinnamon_generic.components.helper import Helper


class TFHelper(Helper):

    def set_seed(
            self,
            seed: int
    ):
        super().set_seed(seed=seed)
        tf.random.set_seed(seed)

        if self.deterministic:
            os.environ['TF_DETERMINISTIC_OPS'] = '1'
            try:
                tf.config.experimental.enable_op_determinism()
            except AttributeError:
                pass

    def clear_status(
            self
    ):
        K.clear_session()

    def limit_gpu_usage(self):
        gpus = tf.config.experimental.list_physical_devices('GPU')

        # avoid other GPUs
        if self.limit_gpu_visibility:
            tf.config.set_visible_devices([gpu for idx, gpu in enumerate(gpus) if idx in self.gpu_indexes], "GPU")

        if gpus:
            try:
                for gpu in gpus:
                    tf.config.experimental.set_memory_growth(gpu, True)
            except RuntimeError as e:
                logging_utility.logger.exception(e)

    def enable_eager_execution(self):
        assert tf.version.VERSION.startswith('2.'), \
            "Tensorflow version is not 2.X! This framework only supports >= 2.0 TF versions"
        tf.config.run_functions_eagerly(self.eager_execution)

    def run(
            self,
            seed: Optional[int] = None
    ) -> Any:
        self.set_seed(seed=seed)
        self.limit_gpu_usage()

        if self.eager_execution:
            self.enable_eager_execution()



from pathlib import Path

from setuptools import setup

readme_path = Path(__file__).absolute().parent.joinpath('README.md')

with readme_path.open('r', encoding='utf-8') as fh:
    long_description = fh.read()

setup(
    name='cinnamon_tf',
    version='0.1',
    author='Federico Ruggeri',
    author_email='federico.ruggeri6@unibo.it',
    description='[Tensorflow Package] A simple high-level framework for research',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://github.com/federicoruggeri/cinnamon_tf',
    project_urls={
        'Bug Tracker': "https://github.com/federicoruggeri/cinnamon_tf/issues"
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent"
    ],
    license='MIT',
    packages=['cinnamon_tf',
              'cinnamon_tf.components',
              'cinnamon_tf.configurations',
              ],
    python_requires=">=3.6"
)
